from typing import Any

EVENT_TYPE = 'event'


def is_event(message: Any) -> bool:
    # Check if this is the new format message
    if isinstance(message, dict) and message.get('type') == EVENT_TYPE:
        payload = message.get('payload')
        return isinstance(payload, dict) and isinstance(payload.get('hash'), str)

    # Otherwise check for old format (just for backward compatibility)
    if not isinstance(message, dict):
        return False

    payload = message.get('payload')
    return (
        isinstance(payload, dict)
        and isinstance(payload.get('event'), dict)
        and isinstance(payload.get('hash'), str)
    )



from dataclasses import dataclass
from typing import TypeVar, Generic

DataT = TypeVar('DataT')


@dataclass
class Return(Generic[DataT]):
    data: DataT


@dataclass
class Retry:
    cause: Exception


RetryResult = Return[DataT] | Retry



from collections.abc import Callable, Coroutine
import time
from typing import TypeVar

from .retry_error import RetryError
from .retry_result import RetryResult, Return
from ..get_randomized_duration import get_randomized_duration


ReturnT = TypeVar('ReturnT')


async def retry_with_backoff(
    tries: int,
    operation: Callable[[], Coroutine[None, None, RetryResult[ReturnT]]],
) -> ReturnT:
    if tries < 1:
        raise ValueError('Tries must be greater than 0')

    retry_error = RetryError()

    for tries_count in range(tries):
        timeout = get_randomized_duration(300, 100) * tries_count

        time.sleep(timeout / 1_000)

        result = await operation()

        if isinstance(result, Return):
            return result.data

        retry_error.append_error(result.cause)

    raise retry_error



from dataclasses import dataclass

from .validate_subject import validate_subject
from .validate_type import validate_type


@dataclass
class EventCandidate:
    source: str
    subject: str
    type: str
    data: dict
    trace_parent: str = None
    trace_state: str = None

    def validate(self) -> None:
        validate_subject(self.subject)
        validate_type(self.type)

    def to_json(self):
        json = {
            'data': self.data,
            'source': self.source,
            'subject': self.subject,
            'type': self.type
        }

        if self.trace_parent is not None:
            json['traceparent'] = self.trace_parent
        if self.trace_state is not None:
            json['tracestate'] = self.trace_state

        return json



from dataclasses import dataclass
from typing import TypeVar

from ...errors.validation_error import ValidationError

Self = TypeVar("Self", bound="EventType")


@dataclass
class EventType:
    event_type: str
    is_phantom: bool
    schema: str | None = None

    @staticmethod
    def parse(unknown_object: dict) -> Self:
        event_type = unknown_object.get('eventType')
        if not isinstance(event_type, str):
            raise ValidationError(
                f"Failed to parse eventType '{event_type}' to str."
            )

        is_phantom = unknown_object.get('isPhantom')
        if not isinstance(is_phantom, bool):
            raise ValidationError(
                f"Failed to parse isPhantom '{is_phantom}' to bool."
            )

        schema = unknown_object.get('schema')
        if schema is not None and not isinstance(schema, str):
            raise ValidationError(
                f"Failed to parse schema '{schema}' to str."
            )

        return EventType(
            event_type=event_type,
            is_phantom=is_phantom,
            schema=schema,
        )

    def __hash__(self):
        return hash((self.event_type, self.is_phantom, self.schema))



from dataclasses import dataclass

from ..lower_bound import LowerBound
from ..upper_bound import UpperBound
from .order import Order
from ...errors.validation_error import ValidationError
from ...event.validate_subject import validate_subject
from .read_from_latest_event import ReadFromLatestEvent


@dataclass
class ReadEventsOptions:
    recursive: bool
    order: Order | None = None
    lower_bound: LowerBound | None = None
    upper_bound: UpperBound | None = None
    from_latest_event: ReadFromLatestEvent | None = None

    def validate(self) -> None:
        # Update validation logic for new object types
        if self.lower_bound is not None and not isinstance(self.lower_bound, LowerBound):
            raise ValidationError(
                'ReadEventOptions are invalid: lower_bound must be a LowerBound object.'
            )

        if self.upper_bound is not None and not isinstance(self.upper_bound, UpperBound):
            raise ValidationError(
                'ReadEventOptions are invalid: upper_bound must be a UpperBound object.'
            )

        if self.from_latest_event is not None:
            if self.lower_bound is not None:
                raise ValidationError(
                    'ReadEventsOptions are invalid: '
                    'lowerBound and fromLatestEvent are mutually exclusive'
                )

            try:
                validate_subject(self.from_latest_event.subject)
            except ValidationError as validation_error:
                raise ValidationError(
                    f'ReadEventsOptions are invalid: '
                    f'from_latest_event.subject: {str(validation_error)}'
                ) from validation_error

            # Add validation for empty type too
            if not self.from_latest_event.type:
                raise ValidationError(
                    'ReadEventsOptions are invalid: '
                    'from_latest_event.type cannot be empty'
                )

    def to_json(self):
        json = {
            'recursive': self.recursive
        }

        if self.order is not None:
            json['order'] = self.order.value

        # Directly use the objects
        if self.lower_bound is not None:
            json['lowerBound'] = {
                'id': str(self.lower_bound.id),  # Ensure ID is a string
                'type': self.lower_bound.type
            }

        if self.upper_bound is not None:
            json['upperBound'] = {
                'id': str(self.upper_bound.id),  # Ensure ID is a string
                'type': self.upper_bound.type
            }

        if self.from_latest_event is not None:
            json['fromLatestEvent'] = self.from_latest_event.to_json()

        return json



from abc import ABC, abstractmethod
from dataclasses import dataclass


class Precondition(ABC):
    @abstractmethod
    def to_json(self):
        raise NotImplementedError()


@dataclass
class IsSubjectPristinePrecondition(Precondition):
    subject: str

    def to_json(self):
        return {
            'type': 'isSubjectPristine',
            'payload': {
                'subject': self.subject
            }
        }


@dataclass
class IsSubjectOnEventIdPrecondition(Precondition):
    subject: str
    event_id: str

    def to_json(self):
        return {
            'type': 'isSubjectOnEventId',
            'payload': {
                'subject': self.subject,
                'eventId': self.event_id
            }
        }



import re

from ..errors.validation_error import ValidationError

TYPE_PATTERN = r'^[0-9A-Za-z_-]{2,}\.(?:[0-9A-Za-z_-]+\.)+[0-9A-Za-z_-]+$'


def validate_type(event_type: str) -> None:
    if re.search(TYPE_PATTERN, event_type) is None:
        raise ValidationError(
            f'Failed to validate type: \'{event_type}\' must be a reverse domain name.'
        )



from .custom_error import CustomError


class InvalidParameterError(CustomError):
    def __init__(self, parameter_name: str, reason: str):
        super().__init__(
            f'Parameter \'{parameter_name}\' is invalid: {reason}')



def join_segments(first: str, *rest: str) -> str:
    first_without_trailing_slash = first.rstrip('/')
    rest_joined = '/'.join([segment.strip('/') for segment in rest])

    return f'{first_without_trailing_slash}/{rest_joined}'



import random


def get_randomized_duration(
    duration_milliseconds: int,
    deviation_milliseconds: int
) -> int:
    return duration_milliseconds \
        - deviation_milliseconds \
        + round(random.random() * deviation_milliseconds * 2)



from http import HTTPStatus
import json

from ...abstract_base_client import AbstractBaseClient
from ...errors.custom_error import CustomError
from ...errors.internal_error import InternalError
from ...errors.invalid_parameter_error import InvalidParameterError
from ...errors.server_error import ServerError
from ...errors.validation_error import ValidationError
from ...event.event_candidate import EventCandidate
from ...event.event_context import EventContext
from ...http_client.response import Response
from .preconditions import Precondition


async def write_events(
    client: AbstractBaseClient,
    event_candidates: list[EventCandidate],
    preconditions: list[Precondition]
) -> list[EventContext]:
    if len(event_candidates) < 1:
        raise InvalidParameterError(
            'event_candidates',
            'event_candidates must contain at least one EventCandidate.'
        )

    for event_candidate in event_candidates:
        try:
            event_candidate.validate()
        except ValidationError as validation_error:
            raise InvalidParameterError(
                'event_candidates', str(validation_error)
            ) from validation_error
        except Exception as other_error:
            raise InternalError(str(other_error)) from other_error

    request_body = json.dumps({
        'events': [event_candidate.to_json() for event_candidate in event_candidates],
        'preconditions': [precondition.to_json() for precondition in preconditions]
    })

    response: Response
    try:
        response = await client.http_client.post(
            path='/api/v1/write-events',
            request_body=request_body,
        )
    except CustomError as custom_error:
        raise custom_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    if response.status_code != HTTPStatus.OK:
        raise ServerError(
            f'Unexpected response status: '
            f'{response.status_code} {HTTPStatus(response.status_code).phrase}.'
        )

    response_data = await response.body.read()
    response_data = bytes.decode(response_data, encoding='utf-8')
    try:
        response_data = json.loads(response_data)
    except json.JSONDecodeError as decode_error:
        raise ServerError(str(decode_error)) from decode_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    if not isinstance(response_data, list):
        raise ServerError(
            f'Failed to parse response \'{response_data}\' to list.')

    result = []
    for unparsed_event_context in response_data:
        try:
            result.append(EventContext.parse(unparsed_event_context))
        except ValidationError as validation_error:
            raise ServerError(str(validation_error)) from validation_error
        except Exception as other_error:
            raise InternalError(str(other_error)) from other_error

    return result



from dataclasses import dataclass

from eventsourcingdb.errors.invalid_parameter_error import InvalidParameterError


@dataclass
class LowerBound:
    id: str
    type: str

    def __post_init__(self):
        if self.type not in {'inclusive', 'exclusive'}:
            raise InvalidParameterError(
                parameter_name="LowerBound",
                reason='type must be either "inclusive" or "exclusive"'
            )
        if int(self.id) < 0:
            raise InvalidParameterError(
                parameter_name='LowerBound',
                reason='id must be non-negative'
            )



import json
from http import HTTPStatus

from ...abstract_base_client import AbstractBaseClient
from ...errors.custom_error import CustomError
from ...errors.internal_error import InternalError
from ...errors.invalid_parameter_error import InvalidParameterError
from ...errors.server_error import ServerError
from ...errors.validation_error import ValidationError
from ...event.validate_type import validate_type
from ...http_client.response import Response


async def register_event_schema(
    client: AbstractBaseClient,
    event_type: str,
    json_schema: str,
) -> None:
    try:
        validate_type(event_type)
    except ValidationError as validation_error:
        raise InvalidParameterError(
            'event_type', str(validation_error)
        ) from validation_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    request_body = json.dumps({
        'eventType': event_type,
        'schema': json_schema,
    })

    response: Response
    try:
        response = await client.http_client.post(
            path='/api/v1/register-event-schema',
            request_body=request_body,
        )
    except CustomError as custom_error:
        raise custom_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    with response:
        if response.status_code != HTTPStatus.OK:
            raise ServerError(
                'Unexpected response status: '
                f'{response.status_code} {HTTPStatus(response.status_code).phrase}'
            )

    return



from typing import Any

SUBJECT_TYPE = 'subject'


def is_subject(message: Any) -> bool:
    if not isinstance(message, dict) or message.get('type') != SUBJECT_TYPE:
        return False

    payload = message.get('payload')

    return isinstance(payload, dict) and isinstance(payload.get('subject'), str)



from dataclasses import dataclass

from eventsourcingdb.errors.invalid_parameter_error import InvalidParameterError


@dataclass
class UpperBound:
    id: str
    type: str

    def __post_init__(self):
        if self.type not in {"inclusive", "exclusive"}:
            raise InvalidParameterError(
                parameter_name="UpperBound",
                reason='type must be either "inclusive" or "exclusive"',
            )
        if int(self.id) < 0:
            raise InvalidParameterError(
                parameter_name="UpperBound", reason="id must be non-negative"
            )



from collections.abc import AsyncGenerator

from .abstract_base_client import AbstractBaseClient
from .client_configuration import ClientConfiguration
from .client_options import ClientOptions
from .event.event_candidate import EventCandidate
from .event.event_context import EventContext
from .handlers.observe_events.observe_events import observe_events
from .handlers.observe_events.observe_events_options import ObserveEventsOptions
from .handlers.read_event_types.event_type import EventType
from .handlers.read_event_types.read_event_types import read_event_types
from .handlers.register_event_schema.register_event_schema import register_event_schema
from .http_client.http_client import HttpClient
from .handlers.ping import ping
from .handlers.read_events import read_events, ReadEventsOptions
from .handlers.read_subjects import read_subjects
from .handlers.store_item import StoreItem
from .handlers.write_events import Precondition, write_events


# pylint: disable=R6007
# Reason: This class explicitly specifies the return type as None
# for better readability. Even though it is not necessary,
# it makes the return type clear without needing to read any
# documentation or code.
class Client(AbstractBaseClient):
    def __init__(
        self,
        base_url: str,
        api_token: str,
        options: ClientOptions = ClientOptions()
    ):
        configuration = ClientConfiguration(
            base_url=base_url,
            timeout_seconds=options.timeout_seconds,
            api_token=api_token,
            protocol_version=options.protocol_version,
            max_tries=options.max_tries
        )

        self.__http_client = HttpClient(configuration)

    async def initialize(self) -> None:
        await self.__http_client.initialize()

    async def close(self) -> None:
        await self.__http_client.close()

    @property
    def http_client(self) -> HttpClient:
        return self.__http_client

    async def ping(self) -> None:
        return await ping(self)

    async def read_subjects(
        self,
        base_subject: str
    ) -> AsyncGenerator[str, None]:
        async for subject in read_subjects(self, base_subject):
            yield subject

    async def read_events(
        self,
        subject: str,
        options: ReadEventsOptions
    ) -> AsyncGenerator[StoreItem, None]:
        async for event in read_events(self, subject, options):
            yield event

    async def read_event_types(self) -> AsyncGenerator[EventType, None]:
        async for event_type in read_event_types(self):
            yield event_type

    async def register_event_schema(self, event_type: str, json_schema: str) -> None:
        await register_event_schema(self, event_type, json_schema)

    async def observe_events(
        self,
        subject: str,
        options: ObserveEventsOptions
    ) -> AsyncGenerator[StoreItem, None]:
        async for event in observe_events(self, subject, options):
            yield event

    async def write_events(
        self,
        event_candidates: list[EventCandidate],
        preconditions: list[Precondition] = None
    ) -> list[EventContext]:
        if preconditions is None:
            preconditions = []
        return await write_events(self, event_candidates, preconditions)



from http import HTTPStatus
import json
from ..abstract_base_client import AbstractBaseClient
from ..errors.server_error import ServerError

# Define constants for response values
OK_RESPONSE = "OK"
STATUS_OK = "ok"

# CloudEvent field names
SPECVERSION_FIELD = "specversion"
TYPE_FIELD = "type"
PING_RECEIVED_TYPE = "io.eventsourcingdb.ping-received"


async def ping(client: AbstractBaseClient) -> None:
    response = await client.http_client.get("/api/v1/ping")
    response_body = bytes.decode(await response.body.read(), encoding="utf-8")

    if response.status_code != HTTPStatus.OK:
        raise ServerError(f"Received unexpected response: {response_body}")

    # Check old format (plain "OK")
    if response_body == OK_RESPONSE:
        return

    try:
        response_json = json.loads(response_body)
    except json.JSONDecodeError as exc:
        raise ServerError(f"Received unexpected response: {response_body}") from exc

    # Check if it's a CloudEvent format (has specversion, type fields)
    if (
        isinstance(response_json, dict)
        and SPECVERSION_FIELD in response_json
        and TYPE_FIELD in response_json
    ):
        if response_json.get(TYPE_FIELD) == PING_RECEIVED_TYPE:
            return

    raise ServerError(f"Received unexpected response: {response_body}")



from abc import ABC, abstractmethod
from .http_client.http_client import HttpClient


class AbstractBaseClient(ABC):
    @property
    @abstractmethod
    def http_client(self) -> HttpClient:
        raise NotImplementedError()



from collections.abc import AsyncGenerator
from http import HTTPStatus

from .event_type import EventType
from .is_event_type import is_event_type
from ..is_stream_error import is_stream_error
from ..parse_raw_message import parse_raw_message
from ...abstract_base_client import AbstractBaseClient
from ...errors.custom_error import CustomError
from ...errors.internal_error import InternalError
from ...errors.server_error import ServerError
from ...errors.validation_error import ValidationError
from ...http_client.response import Response

# pylint: disable=R6007
# Reason: This method explicitly specifies the return type as None
# for better readability. Even though it is not necessary,
# it makes the return type clear without needing to read any
# documentation or code.


async def read_event_types(
    client: AbstractBaseClient,
) -> AsyncGenerator[EventType, None]:
    response: Response
    try:
        response = await client.http_client.post(
            path='/api/v1/read-event-types',
            request_body='',
        )
    except CustomError as custom_error:
        raise custom_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    with response:
        if response.status_code != HTTPStatus.OK:
            raise ServerError(
                'Unexpected response status: '
                f'{response.status_code} {HTTPStatus(response.status_code).phrase}'
            )
        async for raw_message in response.body:
            message = parse_raw_message(raw_message)

            if is_stream_error(message):
                raise ServerError(message['payload']['error'])

            if is_event_type(message):
                event_type: EventType
                try:
                    event_type = EventType.parse(message['payload'])
                except ValidationError as validation_error:
                    raise ServerError(str(validation_error)) from validation_error
                except Exception as other_error:
                    raise InternalError(str(other_error)) from other_error

                yield event_type
                continue

            raise ServerError(
                f'Failed to read event types, an unexpected stream item was received \'{message}\'.'
            )



import json
from collections.abc import AsyncGenerator
from http import HTTPStatus

from ..is_heartbeat import is_heartbeat
from ..is_event import is_event
from ..is_stream_error import is_stream_error
from ..parse_raw_message import parse_raw_message
from ...abstract_base_client import AbstractBaseClient
from ...errors.custom_error import CustomError
from ...errors.internal_error import InternalError
from ...errors.invalid_parameter_error import InvalidParameterError
from ...errors.server_error import ServerError
from ...errors.validation_error import ValidationError
from ...event.event import Event
from ...event.validate_subject import validate_subject
from ..store_item import StoreItem
from .observe_events_options import ObserveEventsOptions
from ...http_client.response import Response

# pylint: disable=R6007
# Reason: This method explicitly specifies the return type as None
# for better readability. Even though it is not necessary,
# it makes the return type clear without needing to read any
# documentation or code.


async def observe_events(
    client: AbstractBaseClient,
    subject: str,
    options: ObserveEventsOptions
) -> AsyncGenerator[StoreItem, None]:
    try:
        validate_subject(subject)
    except ValidationError as validation_error:
        raise InvalidParameterError('subject', str(validation_error)) from validation_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    try:
        options.validate()
    except ValidationError as validation_error:
        raise InvalidParameterError('options', str(validation_error)) from validation_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    request_body = json.dumps({
        'subject': subject,
        'options': options.to_json()
    })

    response: Response
    try:
        response = await client.http_client.post(
            path='/api/v1/observe-events',
            request_body=request_body,
        )
    except CustomError as custom_error:
        raise custom_error
    except Exception as other_error:
        raise InternalError(str(other_error)) from other_error

    with response:
        if response.status_code != HTTPStatus.OK:
            raise ServerError(
                f'Unexpected response status: '
                f'{response.status_code} {HTTPStatus(response.status_code).phrase}'
            )
        async for raw_message in response.body:
            message = parse_raw_message(raw_message)

            if is_heartbeat(message):
                continue

            if is_stream_error(message):
                raise ServerError(f'{message["payload"]["error"]}.')

            if is_event(message):
                event = Event.parse(message['payload'])
                # Add client-side filtering by event ID
                event_id = int(message['payload']['id'])

                if options.lower_bound is not None:
                    # For inclusive, include events with ID >= lower bound
                    if (
                        options.lower_bound.type == 'inclusive' and  # pylint: disable=R2004
                        int(event_id) < int(options.lower_bound.id)
                    ):
                        continue
                    # For exclusive, include events with ID > lower bound
                    if (
                        options.lower_bound.type == 'exclusive' and  # pylint: disable=R2004
                        int(event_id) <= int(options.lower_bound.id)
                    ):
                        continue

                yield StoreItem(event, message['payload']['hash'])
                continue

            raise ServerError(
                f'Failed to read events, an unexpected stream item was received: '
                f'{message}.'
            )



from ...errors.custom_error import CustomError


class RetryError(CustomError):
    def __init__(self):
        self.__errors = []

    def append_error(self, error: Exception):
        super().__init__()
        self.__errors.append(error)

    def __str__(self) -> str:
        return (
            f'Failed operation with {len(self.__errors)} error(s):\n'
            '\n'.join(str(self.__errors))
        )



import asyncio
from collections.abc import Mapping
from http import HTTPStatus

import aiohttp

Headers = Mapping[str, str]


class Response:
    def __init__(self, response: aiohttp.ClientResponse):
        self.__response: aiohttp.ClientResponse = response

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.__response.close()

    @property
    def status_code(self) -> HTTPStatus:
        return HTTPStatus(self.__response.status)

    @property
    def headers(self) -> Headers:
        return self.__response.headers

    @property
    def body(self) -> asyncio.StreamReader:
        return self.__response.content



from typing import Any

ERROR_TYPE = 'error'


def is_stream_error(message: Any) -> bool:
    if not isinstance(message, dict) or message.get('type') != ERROR_TYPE:
        return False

    payload = message.get('payload')

    return isinstance(payload, dict) and isinstance(payload.get('error'), str)



from dataclasses import dataclass

from .if_event_is_missing_during_observe import IfEventIsMissingDuringObserve


@dataclass
class ObserveFromLatestEvent:
    subject: str
    type: str
    if_event_is_missing: IfEventIsMissingDuringObserve

    def to_json(self):
        return {
            'subject': self.subject,
            'type': self.type,
            'ifEventIsMissing': self.if_event_is_missing
        }



import re

from ..errors.validation_error import ValidationError

WORD_PATTERN = '[0-9A-Za-z_-]+'
SUBJECT_PATTERN = f'^/({WORD_PATTERN}/)*({WORD_PATTERN}/?)?$'


def validate_subject(subject: str) -> None:
    if re.search(SUBJECT_PATTERN, subject) is None:
        raise ValidationError(
            f'Failed to validate subject: \'{subject}\' must be an absolute, slash-separated path.'
        )



from datetime import datetime
from typing import TypeVar

from ..errors.validation_error import ValidationError
from .event_context import EventContext

Self = TypeVar("Self", bound="Event")

# pylint: disable=R0917
# Reason: This class is expected to have many parameters
# due to its business context. Splitting it into smaller
# methods would increase cognitive load and make the
# code less readable.


class Event(EventContext):
    def __init__(
        self,
        data: dict,
        source: str,
        subject: str,
        event_type: str,
        spec_version: str,
        event_id: str,
        time: datetime,
        data_content_type: str,
        predecessor_hash: str,
        trace_parent: str = None,
        trace_state: str = None
    ):
        super().__init__(
            source,
            subject,
            event_type,
            spec_version,
            event_id,
            time,
            data_content_type,
            predecessor_hash,
            trace_parent,
            trace_state
        )
        self.data = data

    @staticmethod
    def parse(unknown_object: dict) -> Self:
        event_context = super(Event, Event).parse(unknown_object)

        data = unknown_object.get('data')
        if not isinstance(data, dict):
            raise ValidationError(
                f'Failed to parse data \'{data}\' to object.'
            )

        return Event(
            data,
            event_context.source,
            event_context.subject,
            event_context.type,
            event_context.spec_version,
            event_context.event_id,
            event_context.time,
            event_context.data_content_type,
            event_context.predecessor_hash,
            event_context.trace_parent,
            event_context.trace_state
        )

    def to_json(self):
        json = super().to_json()
        json['data'] = self.data

        return json



from collections import namedtuple
from collections.abc import Callable, Coroutine
from http import HTTPStatus

import aiohttp
from aiohttp import ClientSession

from .response import Response, Headers
from ..client_configuration import ClientConfiguration
from ..errors.client_error import ClientError
from ..errors.custom_error import CustomError
from ..errors.internal_error import InternalError
from ..errors.server_error import ServerError
from ..util import url
from ..util.retry.retry_result import Retry, Return, RetryResult
from ..util.retry.retry_with_backoff import retry_with_backoff
from ..util.retry.retry_error import RetryError

Range = namedtuple('Range', 'lower, upper')


class UninitializedError(CustomError):
    pass


class HttpClient:
    def __init__(
        self,
        client_configuration: ClientConfiguration
    ):
        self.__client_configuration: ClientConfiguration = client_configuration
        self.__session: ClientSession | None = None

    async def initialize(self) -> None:
        self.__session = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                connect=self.__client_configuration.timeout_seconds,
                sock_read=self.__client_configuration.timeout_seconds
            )
        )

    async def close(self):
        if self.__session is not None:
            await self.__session.close()

    async def __execute_request(
        self,
        execute_request: Callable[[], Coroutine[None, None, RetryResult[Response]]]
    ) -> Response:
        try:
            return await retry_with_backoff(
                self.__client_configuration.max_tries,
                execute_request
            )
        except RetryError as retry_error:
            raise ServerError(str(retry_error)) from retry_error
        except CustomError as custom_error:
            raise custom_error
        except aiohttp.ClientError as request_error:
            raise ServerError(str(request_error)) from request_error
        except Exception as other_error:
            raise InternalError(str(other_error)) from other_error

    def __validate_protocol_version(self, http_status_code: int, headers: Headers) -> None:
        if http_status_code != HTTPStatus.UNPROCESSABLE_ENTITY:
            return

        server_protocol_version = headers.get(
            'x-eventsourcingdb-protocol-version'
        )

        if server_protocol_version is None:
            server_protocol_version = 'unknown version'

        raise ClientError(
            f'Protocol version mismatch, server \'{server_protocol_version}\','
            f' client \'{self.__client_configuration.protocol_version}\'.'
        )

    @staticmethod
    async def __get_error_message(response: Response):
        error_message = f'Request failed with status code \'{response.status_code}\''

        # We want to purposefully ignore all errors here, as we're already error handling,
        # and this function just tries to get more information on a best-effort basis.
        # pylint: disable=too-many-try-statements
        try:
            encoded_error_reason = await response.body.read()
            error_reason = encoded_error_reason.decode('utf-8')
            error_message += f" {error_reason}"
        finally:
            pass
        # pylint: enable=too-many-try-statements

        error_message += '.'

        return error_message

    async def __validate_response(
        self,
        response: Response
    ) -> RetryResult[Response]:
        server_failure_range = Range(500, 600)
        if server_failure_range.lower <= response.status_code < server_failure_range.upper:
            return Retry(ServerError(await self.__get_error_message(response)))

        client_failure_range = Range(400, 500)
        if client_failure_range.lower <= response.status_code < client_failure_range.upper:
            raise ClientError(await self.__get_error_message(response))

        self.__validate_protocol_version(response.status_code, response.headers)

        return Return(response)

    def __get_post_request_headers(self) -> dict[str, str]:
        headers = {
            'X-EventSourcingDB-Protocol-Version': self.__client_configuration.protocol_version,
            'Authorization': f'Bearer {self.__client_configuration.api_token}',
            'Content-Type': 'application/json'
        }

        return headers

    async def post(self, path: str, request_body: str) -> Response:
        if self.__session is None:
            raise UninitializedError()

        async def execute_request() -> RetryResult[Response]:
            response = await self.__session.post(
                url.join_segments(
                    self.__client_configuration.base_url,
                    path
                ),
                data=request_body,
                headers=self.__get_post_request_headers(),
            )

            response = Response(response)
            try:
                result = await self.__validate_response(response)
            except Exception as error:
                response.close()
                raise error

            if isinstance(result, Retry):
                response.close()

            return result

        return await self.__execute_request(execute_request)

    def __get_get_request_headers(self, with_authorization: bool) -> dict[str, str]:
        headers = {
            'X-EventSourcingDB-Protocol-Version': self.__client_configuration.protocol_version,
        }

        if with_authorization:
            headers['Authorization'] = f'Bearer {self.__client_configuration.api_token}'

        return headers

    async def get(
        self,
        path: str,
        with_authorization: bool = True,
    ) -> Response:
        if self.__session is None:
            raise UninitializedError()

        async def execute_request() -> RetryResult[Response]:
            response = await self.__session.get(
                url.join_segments(
                    self.__client_configuration.base_url, path),
                headers=self.__get_get_request_headers(with_authorization),
            )

            response = Response(response)
            try:
                result = await self.__validate_response(response)
            except Exception as error:
                response.close()
                raise error

            if isinstance(result, Retry):
                response.close()

            return result

        return await self.__execute_request(execute_request)



from dataclasses import dataclass

from ..event.event_candidate import EventCandidate


# pylint: disable=R0917
# Reason: This class is expected to have many parameters
# due to its business context. Splitting it into smaller
# methods would increase cognitive load and make the
# code less readable.
@dataclass
class Source:
    source: str

    def new_event(
        self,
        subject: str,
        event_type: str,
        data: dict,
        trace_parent: str = None,
        trace_state: str = None
    ) -> EventCandidate:
        return EventCandidate(
            self.source,
            subject,
            event_type,
            data,
            trace_parent,
            trace_state
        )



from .read_events import read_events
from .read_events_options import ReadEventsOptions
from .read_from_latest_event import ReadFromLatestEvent
from .if_event_is_missing_during_read import IfEventIsMissingDuringRead



import json
from typing import Any

from ..errors.server_error import ServerError


def parse_raw_message(raw_message: bytes) -> Any:
    decoded_message: str
    try:
        decoded_message = raw_message.decode('utf8')
    except Exception as error:
        raise ServerError(str(error)) from error

    try:
        return json.loads(decoded_message)
    except Exception as error:
        raise ServerError(str(error)) from error



from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar

from ..errors.internal_error import InternalError
from ..errors.validation_error import ValidationError
from .validate_subject import validate_subject
from .validate_type import validate_type

Self = TypeVar("Self", bound="EventContext")


@dataclass
class EventContext:
    source: str
    subject: str
    type: str
    spec_version: str
    event_id: str
    time: datetime
    data_content_type: str
    predecessor_hash: str
    trace_parent: str = None
    trace_state: str = None

    @staticmethod
    def parse(unknown_object: dict) -> Self:
        source = unknown_object.get('source')
        if not isinstance(source, str):
            raise ValidationError(
                f'Failed to parse source \'{source}\' to string.')

        subject = unknown_object.get('subject')
        if not isinstance(subject, str):
            raise ValidationError(
                f'Failed to parse subject \'{subject}\' to string.')
        validate_subject(subject)

        event_type = unknown_object.get('type')
        if not isinstance(event_type, str):
            raise ValidationError(
                f'Failed to parse event_type \'{event_type}\' to string.')
        validate_type(event_type)

        spec_version = unknown_object.get('specversion')
        if not isinstance(spec_version, str):
            raise ValidationError(
                f'Failed to parse spec_version \'{spec_version}\' to string.')

        event_id = unknown_object.get('id')
        if not isinstance(event_id, str):
            raise ValidationError(
                f'Failed to parse event_id \'{event_id}\' to string.')

        time = unknown_object.get('time')
        if not isinstance(time, str):
            raise ValidationError(
                f'Failed to parse time \'{time}\' to datetime.')

        rest, sub_seconds = time.split('.')
        sub_seconds = f'{sub_seconds[:6]:06}'
        try:
            time = datetime.fromisoformat(f'{rest}.{sub_seconds}')
        except ValueError as value_error:
            raise ValidationError(
                f'Failed to parse time \'{time}\' to datetime.') from value_error
        except Exception as other_error:
            raise InternalError(str(other_error)) from other_error

        data_content_type = unknown_object.get('datacontenttype')
        if not isinstance(data_content_type, str):
            raise ValidationError(
                f'Failed to parse data_content_type \'{data_content_type}\' to string.')

        predecessor_hash = unknown_object.get('predecessorhash')
        if not isinstance(predecessor_hash, str):
            raise ValidationError(
                f'Failed to parse predecessor_hash \'{predecessor_hash}\' to string.')

        trace_parent = unknown_object.get('traceparent')
        if trace_parent is not None and not isinstance(trace_parent, str):
            raise ValidationError(
                f'Failed to parse trace_parent \'{trace_parent}\' to string.')

        trace_state = unknown_object.get('tracestate')
        if trace_state is not None and not isinstance(trace_state, str):
            raise ValidationError(
                f'Failed to parse trace_state \'{trace_state}\' to string.')

        return EventContext(
            source=source,
            subject=subject,
            type=event_type,
            spec_version=spec_version,
            event_id=event_id,
            time=time,
            data_content_type=data_content_type,
            predecessor_hash=predecessor_hash,
            trace_parent=trace_parent,
            trace_state=trace_state,
        )

    def to_json(self):
        json = {
            'specversion': self.spec_version,
            'id': self.event_id,
            'time': self.time.isoformat(sep='T'),
            'source': self.source,
            'subject': self.subject,
            'type': self.type,
            'datacontenttype': self.data_content_type,
            'predecessorhash': self.predecessor_hash
        }

        if self.trace_parent is not None:
            json['traceparent'] = self.trace_parent
        if self.trace_state is not None:
            json['tracestate'] = self.trace_state

        return json



from .observe_events import observe_events
from .observe_events_options import ObserveEventsOptions
from .observe_from_latest_event import ObserveFromLatestEvent
from .if_event_is_missing_during_observe import IfEventIsMissingDuringObserve



from typing import Any

EVENT_TYPE_TYPE = 'eventType'


def is_event_type(message: Any) -> bool:
    if not isinstance(message, dict) or message.get('type') != EVENT_TYPE_TYPE:
        return False

    payload = message.get('payload')

    return isinstance(payload, dict)
